///////////////////////////////////////////////////////////////////////////////
//
// edimpro1.js
//
// 1st-order Markov-based improvisation structure generator.
//
// * $$ Last modified: 11:04:04 Mon Mar 21 2011 CET
//
// Michael Edwards ~ m@michael-edwards.org ~ 1.3.10
//
///////////////////////////////////////////////////////////////////////////////
// A piece can be for as many players as we want but there are always
// 4-9 soloists and the rest are the 'tutti'.  if we indicate 'tutti'
// then everyone plays: soloists not playing do so in a supporting
// role.

// what is organised/displayed: section length countdown; players; soloist, if
// any; tutti if applicable; harmony; energy; dynamic 

// always display the soloists of the next section too so people can prepare.

// Default duration is 20 mins.  User can select a duration of 5-60 mins; the
// selected duration will then determine a scaler to the basic unit duration of
// 10 seconds, but this is limited to min 0.5 and max 2.0

// Section lengths are by default 10 second multiples of 2,3,5,13,21 (surprise
// surprise!) but these become longer relative to the ratio of selected
// duration to default duration.

// if options set (true by default):
// always make tutti false at beg and true at end?
// don't allow the higher durations to be dedicated to < 4 players
// use fixed randomness (not default)
////////////////////////////////////////////////////////////////////////////////
// todo:

// change names page so we have text fields for all 9 players

// don't allow solos at beg and end

// add solo trading in middle

// add duration of next section

// add pause before first section starts

// add more/better curves

// done: forced linking places so no complete changes of personnel between
// sections 

// done: turn countdown into 'progress bar'

// done with option: only show info around change points; otherwise blank screen

// done with subsets: remove harmony (or reduce to a few notes: use kill
// subsets)  

// done: remove energy and dynamic and section change blink

// done: no elapsed time but show progress bar through
// whole piece under a global energy curve

////////////////////////////////////////////////////////////////////////////////


// NB I've embedded edimpro1.maxpat in main patch so edited names would be
// saved. Edit that patch by right-clicking on it and selecting Object-New
// View; when done, just close it but be sure to save the main patch.

// text colour: .79 .33 91
// nb purple bgcolor rgb: .19 .02 .19

var edimproData = {
  //  for debugging speed up (> 1) the sections but for performance this has
  // to  be 1 
  speedup : 1,
  blankDelay : 6,
  defaultDuration : 20.0, // minutes
  defaultNumPlayers : 9,
  // what the durations are multiplied by (seconds)
  unitDuration : 10.0,
  // the selected duration is used to scale durations based on the default
  // durations: this could lead to very short or long sections so limit these.
  minScaler : 0.75,
  maxScaler : 2.0,
  // what % chance is there of a section including the tutti
  tuttiChance : 50.0,
  // what % chance is there of a section having a single soloist selected from
  // the current players.
  soloistChance : 25.0,
  // if we add or subtract players here we have to update the players markov
  // rules below
  playerNames : new Array("player1", "player2", "player3", "player4", 
                          "player5", "player6", "player7", "player8", 
                          "player9"),
  // here we define what a 'small group is' i.e anything <= to this, so
  // we can restrict durations if the user wishes.
  smallGroupThreshold : 3,
  // see sets.lsp for conversion from note names/octaves to these midi notes
  // again: if we add/remove chords from here we have to update the changes
  // rules also.
  harmonies : [ [ 62,65,70,76,83, ],
                [ 61,64,72,76,83, ],
                [ 58,63,80, ],
                [ 61,66,70,79,84, ],
                [ 36,41,64,69, ],
                [ 42,52,60,63,69, ],
                [ 32,40,61,74, ],
                [ 49,53,59,64,71, ],
                [ 45,53,71,79, ],
                [ 48,52,69,75, ],
                [ 50,54,68,73, ],
                [ 51,55,65,73, ], ],
  // the following slots are all MDEMarkov rules 
  // harmonic sequences: equal chances for now of a particular chord moving to
  // any of its possible followers  
  changes :   [ [ 1, 2,.5, 4,.5 ],
                [ 2, 3,.5, 3,.5, 9,.5, 1,.5 ],
                [ 3, 4,.5, 10,.5 ],
                [ 4, 1,.5, 2,.5, 3,.5, 9,.5 ],
                [ 5, 6,.5, 6,.5, 12,.5, 7,.5, 12,.5 ],
                [ 6, 7,.5, 5,.5, 12,.5, 8,.5 ],
                [ 7, 8,.5, 5,.5, 8,.5, 6,.5, 9,.5 ],
                [ 8, 9,.5, 11,.5, 6,.5, 10,.5 ],
                [ 9, 10,.5, 7,.5, 3,.5, 8,.5 ],
                [ 10, 11,.5, 5,.5, 8,.5, 3,.5, 9,.5 ],
                [ 11, 12,.5, 8,.5, 12,.5 ],
                [ 12, 11,.5, 5,.5, 6,.5 ] ],
  // I think the duration of 1 is too small so start at 2
  durations : [ [ 2, 3,.5, 5,.3, 13,.1 ],
                [ 3, 21,.1, 2,.4, 5,.4, 13,.05],
                [ 5, 3,.4, 8,.4, 5,.3, ],
                [ 8, 5,.4, 13,.1, 21,.03,],
                [ 13, 2,.4, 8,.2, 21,.05 ],
                [ 21, 2,.5, 5,.2, 8, .1] ],
  // how many players there will be
  players :   [ [ 1, 2,.3, 3,.3, 5, .1 ],
                [ 2, 4,.3, 6,.3, 8,.1  ],
                [ 3, 3,.5, 7,.1, 9,.1  ],
                [ 4, 2,.5, 8,.2, 1,.1, 7,.05],
                [ 5, 1,.5, 2,.3, 9,.05,  ],
                [ 6, 3,.5, 1,.1, 2,.3, 8,.05  ],
                [ 7, 1,.2, 2,.2, 3,.5  ],
                [ 8, 1,.3, 2,.2, 3,.3  ],
                [ 9, 4,.1, 1,.1, 3,.3  ] ],
  // old: we replaced these two with an energy/dynamic curve
  energy :    [ [ 1, 6,.3, 3,.3, 5, .1 ],
                [ 2, 4,.3, 5,.2, 8,.1  ],
                [ 3, 1,.3, 5,.3, 9,.1  ],
                [ 4, 7,.3, 2,.2, 1,.1, 10,.05],
                [ 5, 1,.05, 3,.2, 7,.3, 8,.05, ],
                [ 6, 2,.5, 1,.1, 4,.3, 10,.05, 1,.03  ],
                [ 7, 5,.2, 9,.2, 3,.1, 1,.1 ],
                [ 8, 10,.3, 5,.3, 3,.3  ],
                [ 9, 4,.1, 1,.1, 3,.3, 10,.1  ],
                [ 10, 7,.3, 5,.3, 3,.2  ] ],
  // for the mo, just try the same rules as energy...
  dynamic :   [ [ 1, 6,.3, 3,.3, 5, .1 ],
                [ 2, 4,.3, 5,.2, 8,.1  ],
                [ 3, 1,.3, 5,.3, 9,.1  ],
                [ 4, 7,.3, 2,.2, 1,.1, 10,.05],
                [ 5, 1,.05, 3,.2, 7,.3, 8,.05, ],
                [ 6, 2,.5, 1,.1, 4,.3, 10,.05, 1,.03  ],
                [ 7, 5,.2, 9,.2, 3,.1, 1,.1 ],
                [ 8, 10,.3, 5,.3, 3,.3  ],
                [ 9, 4,.1, 1,.1, 3,.3, 10,.1  ],
                [ 10, 7,.3, 5,.3, 3,.2  ] ]
}
// a global to allow us to send messages to receive objects in the max patch.
var edimproGlobalMessenger = new Global("edimproGlobalMessenger");
// our piece object
var edimproPiece = new improvPiece();
////////////////////////////////////////////////////////////////////////////////
// Global functions
function setName(name, number) {
  if (!name)
    name = "player" + number;
  if (number <= edimproData.playerNames.length)
    edimproData.playerNames[number - 1] = name;
  else post("\nsetName: incorrect reference (should be 1-9): " + number);
}
function blankDelay(secs) {
  edimproData.blankDelay = secs;
}
function resetNames() {
  for (i = 1; i <= edimproData.playerNames.length; ++i)
    setName(0, i);
}
function printNames() {
  for (i = 0; i < edimproData.playerNames.length; ++i)
    post("\nPlayer" + (i + 1) + ": " + edimproData.playerNames[i]);
}
function showHarmony(num) { // 1-based
  var h;
  if (num < 0)
    h = ["chord", 0, 0];
  else h = edimproPiece.edimproHarmoniesNSlider[num - 1];
  edimproSend("edimproHarmony", h);
}
function duration(dur) {
  edimproPiece.setDuration(dur);
}
function numPlayers(num) {
  edimproPiece.setNumPlayers(num);
}
function stop() {
  edimproPiece.gameOver();
}
function newEnv() {
  edimproPiece.newEnvelope();
}
function print() {
  edimproPiece.print();
}
function start() {
  edimproPiece.init();
  edimproPiece.print();
  edimproPiece.countDownTask.interval = 1000;
  edimproPiece.nextSection();
}
function countdown() {
  edimproPiece.countdown();
}
function options(openingTutti, finalTutti, shorterSmallGroups, fixedRandomness,
                 blankScreen)
{
  edimproPiece.openingTutti = openingTutti;
  edimproPiece.finalTutti = finalTutti;
  edimproPiece.shorterSmallGroups = shorterSmallGroups;
  edimproPiece.fixedRandomness = fixedRandomness;
  edimproPiece.blankScreen = blankScreen;
}
function erasePlayers(playerReceiver, soloistReceiver, tuttiReceiver) {
  for (i = 0; i < edimproData.players.length; ++i)
    edimproDisplayPlayer(i, 1, playerReceiver);
  edimproDisplaySoloist(-1, soloistReceiver);
  edimproDisplayTutti(0, tuttiReceiver);
}
function eraseNextPlayers() {
  erasePlayers("2edimproPlayer", "2edimproSoloist", "2edimproTutti"); 
}
function nextSection() {
  edimproPiece.nextSection();
}
////////////////////////////////////////////////////////////////////////////////
// Local functions
edimproSend.local = 1;
function edimproSend(receiver, message) {
  edimproGlobalMessenger.temp = message;
  edimproGlobalMessenger.sendnamed(receiver, "temp");
}
// to set the text in a label (umenu) we need to send a list
edimproLabel.local = 1;
function edimproLabel(text) {
  return [ "setitem", 0, text];
}
edimproDisplayPlayer.local = 1;
function edimproDisplayPlayer(num, erase, receiver) { // 0 based
  if (!receiver)
    receiver = "1edimproPlayer";
  if (erase)
    edimproSend(receiver + (1+num), edimproLabel(""));
  else edimproSend(receiver + (1+num),
                   edimproLabel(edimproData.playerNames[num]));
}
edimproDisplaySoloist.local = 1;
function edimproDisplaySoloist(num, receiver) { // 0 based
  if (!receiver)
    receiver = "1edimproSoloist";
  if (num >= 0)
    edimproSend(receiver, 
                edimproLabel("solo: " + edimproData.playerNames[num]));
  else edimproSend(receiver, edimproLabel(""));
}
edimproDisplayTutti.local = 1;
function edimproDisplayTutti(val, receiver) { // 0=erase 1=display
  if (!receiver)
    receiver = "1edimproTutti";
  if (val)
    edimproSend(receiver, edimproLabel("TUTTI"));
  else edimproSend(receiver, edimproLabel(""));
}
sectionPerCent.local = 1;
function sectionPerCent(elapsed, duration) {
  // ceiling: so there's more chance of getting to the end of the progress bar
  return Math.ceil(100.0 * (elapsed / duration));
}
////////////////////////////////////////////////////////////////////////////////
// Constructor. players is an array of indices into the player names; EOF is
// indicated by -1
function improvSection(players, soloist, tutti, duration, harmony, energy, 
                       dynamic, startTime, endTime) {
  var i;
  // an array of indices into edimproData.playerNames
  this.players = new Array(players.length);
  for (i = 0; i < players.length; ++i)
    this.players[i] = players[i]; 
  this.soloist = soloist; // an index into edimproData.playerNames
  this.tutti = tutti; // 1 or 0
  this.duration = duration; // seconds
  // an index (actually needs -1 first...) into edimproHarmonies which is an
  // array of an array of midi note numbers
  this.harmony = harmony;
  this.energy = energy; // a number from 0-10
  this.dynamic = dynamic; // a number from 0-10
  this.startTime = startTime;
  this.endTime = endTime;
  this.displayHarmony = function() {
    showHarmony(this.harmony); // 1-based
  }
  this.displayPlayers = function
    (playerReceiver, soloistReceiver, tuttiReceiver) {
    var i;
    erasePlayers(playerReceiver, soloistReceiver, tuttiReceiver);
    for (i = 0; i < this.players.length; ++i)
      edimproDisplayPlayer(this.players[i], 0, playerReceiver);
    if (this.soloist >= 0)
      edimproDisplaySoloist(this.soloist, soloistReceiver);
    edimproDisplayTutti(this.tutti, tuttiReceiver);
  }
  this.print = function() {
    post("\n" + this.startTime + "->" + this.endTime + " " + this.duration +
         " [" + this.players + "] solo: " + this.soloist + " tutti: " +
         this.tutti);
  }
}
////////////////////////////////////////////////////////////////////////////////
// Constructor
function improvPiece() {
  // the section and time display scheduling tasks
  this.sectionTask = new Task(nextSection, this);
  this.countDownTask = new Task(countdown, this);
  // use pseudoRandom numbers (not fixed) by default.
  this.fixedRandomness = 0;
  this.blankScreen = 0;
  // Our fixed/non-fixed sequence random number generator
  this.ranGen = new MDERandomPerCent(this.fixedRandomness);
  this.durations = new MDEMarkov(edimproData.durations, this.ranGen);
  // how many players we'll have
  this.players = new MDEMarkov(edimproData.players, this.ranGen);
  this.energy = new MDEMarkov(edimproData.energy, this.ranGen);
  this.dynamic = new MDEMarkov(edimproData.dynamic, this.ranGen);
  this.changes = new MDEMarkov(edimproData.changes, this.ranGen);
  this.tutti = false;
  this.numPlayers = edimproData.defaultNumPlayers; // this is user settable
  this.runningTotal = 0.0;
  this.durationScaler = 1.0;
  // duration of whole piece
  this.durationSecs = edimproData.defaultDuration * 60.0;
  this.soloist = -1;
  this.currentPlayers = new Array(edimproData.players.length);
  this.sections = new Array();
  this.sectionDur = 0;
  this.currentSection = 0;
  // how many seconds remaining in section
  this.remaining = 0;
  this.elapsedSection = 0;
  // elapsed seconds in whole piece
  this.elapsed = 0;
  // the following are GUI options
  this.openingTutti = 1;
  this.finalTutti = 1;
  this.shorterSmallGroups = 1;
  // if the user option is selected, we'll restrict smaller groups to
  // durations less than the second highest (generally 13 i.e. 2mins10)
  this.nextHighestDuration = 
    this.durations.keys[this.durations.keys.length - 2];
  this.numCurrentPlayers = 0;
  // to display the notes, the nslider object needs note/velocity pairs so
  // invent the latter in here down below.
  this.edimproHarmoniesNSlider = new Array(edimproData.harmonies.length);
  // the nslider object needs note/velocity pairs so invent the latter here
  var i, j;
  for (i = 0; i < edimproData.harmonies.length; ++i) {
    var current = edimproData.harmonies[i];
    current.sort(function(a, b) { return a - b; }); 
    var len = current.length * 2;
    this.edimproHarmoniesNSlider[i] = new Array(len);
    for (j = 0; j < len; j += 2) {
      this.edimproHarmoniesNSlider[i][j] = current[j / 2];
      this.edimproHarmoniesNSlider[i][j+1] = 80; // velocity
    }
    this.edimproHarmoniesNSlider[i].unshift("chord");
  }
  edimproSend("edimproShow", 1);
  edimproSend("edimproMessage", edimproLabel(""));
  this.countdown = function() {
    // does JS do integer arithmetic? best make sure with floor
    var eMins = Math.floor(this.elapsed / 60.0); 
    var eSecs = this.elapsed % 60;
    var percent = sectionPerCent(this.elapsedSection++, this.sectionDur);
    edimproSend("edimproCountDown", this.remaining--)
    edimproSend("edimproElapsedMins", eMins)
    edimproSend("edimproElapsedSecs", eSecs)
    edimproSend("edimproSectionPerCent", percent)
    var totalPercent = sectionPerCent(this.elapsed++, this.durationSecs);
    edimproSend("edimproTotalPerCent", totalPercent)
    if (this.blankScreen) {
      if (this.remaining == edimproData.blankDelay)
        edimproSend("edimproShow", 1);
      else if (this.elapsedSection == edimproData.blankDelay
               // only blank screen if the section is long enough
               && this.sectionDur > (edimproData.blankDelay * 3))
        edimproSend("edimproShow", 0);
    }
  }
  // this is called each time the new section starts.
  this.nextSection = function() {
    // shouldn't be necessary but let's make sure...
    edimproSend("edimproShow", 1);
    edimproSend("edimproSectionStart", this.currentSection);
    if (this.currentSection < this.sections.length) {
      var section = this.sections[this.currentSection++];    
      var nextSection = this.sections[this.currentSection];
      // all of our section objects have already been generated
      if (section) {
        section.displayPlayers(); // current players
        section.displayHarmony();
        edimproSend("edimproEnergy", section.energy);
        edimproSend("edimproDynamic", section.dynamic);
        edimproSend("edimproSectionDur", section.duration);
        if (nextSection)
          // remember this function sets "1edimproPlayer" etc. by default
          nextSection.displayPlayers("2edimproPlayer", "2edimproSoloist",
                                     "2edimproTutti"); 
        else eraseNextPlayers();
        this.remaining = section.duration;
        this.sectionDur = section.duration;
        this.elapsedSection = 0;
      }
      // speed things up here for testing purposes
      this.sectionTask.schedule((section.duration * 1000) / 
                           edimproData.speedup);
      this.countDownTask.cancel();
      this.countDownTask.repeat(section.duration);  
    }
    else this.gameOver(true);
  }
  this.gameOver = function(hide) {
    // i.e. the piece has run, rather than us being at init
    if (this.elapsed && hide) { 
      edimproSend("edimproShow", 0);
      edimproSend("edimproMessage", edimproLabel("game over"));
    }
    erasePlayers();
    eraseNextPlayers();
    this.sectionTask.cancel();
    this.countDownTask.cancel();
    showHarmony(-1);
    edimproSend("edimproCountDown", 0)
    edimproSend("edimproSectionPerCent", 100)
    edimproSend("edimproTotalPerCent", 100)
    edimproSend("edimproEnvelope", "clear");
    edimproSend("edimproSectionDur", 0);
  }
  this.print = function() {
    var i;
    for (i = 0; i < this.sections.length; ++i) {
      section = this.sections[i].print();
    }
  }
  this.getSection = function() {
    var startTime = this.runningTotal;
    this.selectPlayers(); // this sets tutti and soloist too
    // this updates runningTotal hence startTime saved above
    var dur = this.nextDuration(); // -1 if we've finished
    var harmony = this.changes.next();
    var i;
    var energy = this.energy.next();
    var dynamic = this.dynamic.next();
    return (dur < 0) ? null : 
    new improvSection(this.currentPlayers, this.soloist, this.tutti, dur, 
                      harmony, energy, dynamic, startTime,
                      this.runningTotal);
  }
  this.init = function () {
    edimproSend("edimproMessage", edimproLabel(""));
    this.sectionTask.cancel();
    this.countDownTask.cancel();
    showHarmony(-1);
    erasePlayers();
    eraseNextPlayers();
    this.ranGen.reset(this.fixedRandomness);
    this.durations.seed(5);
    this.energy.seed(5);
    this.dynamic.seed(5);
    this.players.seed(5);
    this.changes.seed(1);
    this.currentSection = 0;
    this.elapsed = 0;
    this.runningTotal = 0.0;
    this.tutti = false;
    this.newEnvelope();
    // must delete any previous sections so we stop at the right time
    this.sections = new Array();
    var i = 0;
    var happy = 1;
    while (happy) {
      var section = this.getSection();
      if (section)
        this.sections[i++] = section;
      else happy = 0;
    }
    // the following imply that if we have randomly selected opening and
    // closing tuttis anyway then these options won't change anything
    if (this.openingTutti)
      this.sections[0].tutti = true;
    if (this.finalTutti)
      this.sections[this.sections.length - 1].tutti = true;
  }
  this.deletePlayers = function() {
    var i;
    for (i = 0; i < this.currentPlayers.length; ++i)
      this.currentPlayers[i] = -1;
  }
  //***************************************************
  // though we get the number of players from a markov, the actual players
  // selected will be random. so this fills this.currentPlayers with integer
  // indices into the playerNames array 
  this.selectPlayers = function() {
    var num; 
    var i;
    var noNum = 1;
    while (noNum) {
      // keep getting the next number of players from our markov until we've
      // got a number <= the number of players.  NB this is preferable to
      // simply moduloing as at least we're still using the markov rules
      num = this.players.next(); 
      // init calls getSection in a loop and that calls selectPlayers so we
      // know that this.numCurrentPlayers is currently the number in the last
      // section: making sure this isn't the same is the easiest way of making
      // sure no two sections have the same players (although we could just end
      // up with the same +/- one or more).
      if (num <= this.numPlayers && num != this.numCurrentPlayers)
        noNum = 0;
    }
    // 5.11.10: pick a random player from the last section to play in this
    // section too 
    var linkPlayer = -1;
    if (this.numCurrentPlayers > 1) {
      var max = this.numCurrentPlayers - 1;
      // if we have a soloist it's counter in the number of players but it's
      // not in the array.
      if (this.soloist)
        --max;
      linkPlayer = this.currentPlayers[between(0, max, 1, this.ranGen, null)];
    }
    else if (this.soloist)
      linkPlayer = this.soloist;
    // just initialise to the link player in case that's the loop below doesn't
    // run (num would be 1...)
    var player = linkPlayer;
    post(num + ": " + linkPlayer + "\n");
    // if (!linkPlayer) linkPlayer = -1;
    this.deletePlayers();
    this.currentPlayers[0] = linkPlayer; 
    this.numCurrentPlayers = num;
    // if last section wasn't a tutti, see if this one is
    if (!this.tutti)
      this.tutti = this.ranGen.random() <= edimproData.tuttiChance;
    else this.tutti = false;
    var lastSoloist = this.soloist;
    this.soloist = -1; // no soloist yet
    // randomly select the players but with no repeats!
    for (i = linkPlayer == -1 ? 0 : 1; i < num; ++i) {
      player = between(0, this.numPlayers - 1, 1, this.ranGen, 
                       this.currentPlayers);
      // an edimproData.playerNames array index
      this.currentPlayers[i] = player; 
    }
    // if we select a soloist this time just make it the last player we
    // selected; or if no soloist but we only selected 1 player it's
    // effectively the same
    if (num == 1 || this.ranGen.random() <= edimproData.soloistChance) {
      this.soloist = player; 
      --num;
    }
    // don't have same soloist twice
    if (this.soloist >= 0 && this.soloist == lastSoloist) {
      this.soloist = this.currentPlayers[num - 1];
      this.currentPlayers[num - 1] = lastSoloist;
    }
    // set EOF (-1) to indicate where to stop reading in the array 
    this.currentPlayers[num] = -1;
  } //***************************************************
  this.setDuration = function(duration) { // 5-60 minutes
    this.durationScaler = duration / edimproData.defaultDuration;
    if (this.durationScaler < edimproData.minScaler)
      this.durationScaler = edimproData.minScaler;
    else if (this.durationScaler > edimproData.maxScaler)
      this.durationScaler = edimproData.maxScaler;
    this.durationSecs = duration * 60.0;
  }
  this.setNumPlayers = function(numPlayers) { // 4-9
    this.numPlayers = numPlayers;
  }
  // returns -1 if we're finished.
  // this is called in getSection after selectPlayers so we know the
  // number of players we'll have: this.currentPlayers.length
  this.nextDuration = function () {
    var result = -1;
    var dur = this.durations.next();
    // shorter sections with < 4 players?
    if (this.shorterSmallGroups && this.numCurrentPlayers <=
        edimproData.smallGroupThreshold) {
      var notHappy = 1;
      while (notHappy) {
        if (dur < this.nextHighestDuration)
          notHappy = 0;
        else dur = this.durations.next();
      }
    }
    var scaledDur = dur * edimproData.unitDuration * this.durationScaler;
    if (this.runningTotal < this.durationSecs)
      result = scaledDur;
    this.runningTotal += scaledDur;
    return result;
  }
  this.newEnvelope = function () {
    var envelopes = [ [ 0., 0., 116.483513, 0.356383, 164.835159, 0.31383, 
                        312.087921, 0.595745, 507.692322, 0.558511, 534.065918, 
                        0.164894, 661.538452, 0.292553, 698.901123, 1., 
                        830.769226, 1., 927.472534, 0.515957, 936.263733, 1., 
                        1000., 1., ],
                      [ 0., 0.781915, 96.7033, 0.781915, 156.043961, 0.489362,
                        208.791214, 0.861702, 404.395599, 0.856383, 432.967041,
                        0.180851, 512.087891, 0.180851, 580.219788, 0.398936, 
                        701.098877, 0.06383, 727.472534, 0.505319, 813.186829,
                        0.5, 859.340637, 0.744681, 909.890137, 0.739362,
                        929.670349, 1., 969.230774, 0.058511, 1000., 0.06383, ],
                      [ 0., 0.473404, 72.527473, 0.553191, 140.659348, 0.68617,
                        171.428574, 0.398936, 268.131866, 0.170213, 274.725281,
                        0.287234, 338.461548, 0.287234, 393.406586, 0.531915,
                        402.197815, 0.340426, 496.703308, 0.670213, 560.439575,
                        0.478723, 582.417603, 0.335106, 600., 0.925532, 
                        676.923096, 0.914894, 745.054932, 0.324468, 747.252747,
                        0.771277, 835.164856, 0.765957, 874.725281, 0.196809,
                        918.681335, 0.202128, 956.043945, 1., 1000., 1., ],
                      [ 0., 0., 125.274727, 0.085106, 182.417587, 0.175532, 
                        206.593399, 0.867021, 276.923065, 0.835106, 353.846161,
                        0.324468, 406.593414, 0.324468, 494.505493, 0.452128,
                        516.483521, 0.308511, 553.84613, 0.648936, 659.340637,
                        0.675532, 685.714294, 0.292553, 859.340637, 0.292553,
                        879.12085, 0.106383, 934.065918, 0.106383, 969.230774,
                        0.441489, 1000., 0.457447, ]
                      ];
    // random envelope
    var env = envelopes[between(0, envelopes.length - 1, 1, this.ranGen)];
    edimproSend("edimproEnvelope", "clear");
    for (i= 0; i < env.length; i+= 2) {
      edimproSend("edimproEnvelope", [ env[i], env[i + 1] ]);
    }
  }
}
////////////////////////////////////////////////////////////////////////////////
// Constructor
// e.g. [ [ 1, 2, .6, 3, .3, 4 .1 ], [ 2, 1, .5, 2, .1, 3, .3, 4 .1 ], ...
// assumes all rule keys are numbers only.
function MDEMarkov(rules, randomNumberObject) {
  this.ranGen = randomNumberObject;
  this.rulesets = new Array(rules.length);
  this.current = NaN;
  this.keys = new Array(rules.length);
  var i;
  for (i = 0; i < rules.length; ++i) {
    var rule = new MDEMarkovRule(rules[i]);
    this.rulesets[i] = rule;
    this.keys[i] = rule.key;
  }
  // sorts alphanumerically by default...
  this.keys.sort(function(a, b) { return a - b; }); 
  //*********************
  this.seed = function(s) {
    this.current = s;
  }
  //***************************
  // call with a seed the first time and with NaN thereafter, or set seed
  // separately 
  this.next = function(key) {
    var i;
    var next = NaN;
    var currentRule;
    if (isNaN(key)) {
      if (isNaN(this.current)) {
        post("MDEMarkovRule.next(): must be called with a seed " +
             "the first time.");
      }
      key = this.current;
    }
    for (i = 0; i < this.rulesets.length; i++) {
      currentRule = this.rulesets[i];
      if (currentRule.key == key) {
        next = currentRule.get(this.ranGen);
        break;
      }
    }
    if (isNaN(next)) {
      post("\nMDEMarkovRule.next(): can't find result for " + key);
    }
    else {
      this.current = next;
    }
    return next;
  }
}
///////////////////////////////////////////////////////////////////////////////
// Constructor
// e.g. [ 1, 2, .6, 3, .3, 4 .1 ]
// where 1 is the key and then it's result / chance pairs
function MDEMarkovRule(rule) { 
  this.key = rule[0];
  this.ruleset = new Array();
  var i, j, c;
  var total = 0.0;
  var scaler;
  for (i = 1; i < rule.length - 1; i += 2) {
    j = i + 1;
    this.ruleset[i - 1] = rule[i]; // we don't include the key in our ruleset
    c = rule[j];
    total += c;
    this.ruleset[i] = total;
  }
  scaler = 100.0 / total;
  // now scale all our chances to be 0-100
  for (i = 1; i < this.ruleset.length; i += 2) {
    this.ruleset[i] *= scaler;
  }
  //*****************
  this.get = function(randomObject) {
    // our global random number generator
    var ran = randomObject.random(); 
    var result = NaN;
    for (i = 1; i < this.ruleset.length; i += 2) {
      if (ran < this.ruleset[i]) {
        result = this.ruleset[i - 1];
        break;
      }
    }
    return result;
  }
}
///////////////////////////////////////////////////////////////////////////////
// return a random number between the two arguments (inclusive), rounding if
// desired. 
// 'avoiding' should be either 0 (meaning any number can be returned) or an
// array of numbers to avoid.
between.local = 1;
function between(lower, upper, round, randomObject, avoiding) { 
  var maxtries = 1000;
  var result;
  var happy = 1;
  var i;
  while (happy && happy < maxtries) {   // e.g. max 1000 attempts
    var gotit = 0;
    result = lower + ((upper - lower) * (randomObject.random() / 100.0));
    if (round)
      result = Math.round(result);
    if (!avoiding)
      happy = 0;
    else {
      for (i = 0; i < avoiding.length && !gotit; ++i) {
        gotit = avoiding[i] == result;
      }
    }
    if (!gotit)
      happy = 0;
    else ++happy;
  }
  if (happy == maxtries)
    post("\nCouldn't get a unique result: " + lower + " " + upper + 
         " avoiding " + avoiding);
  return result;
}
// Constructor
// if fixed==1, we'll use the fixed set of random numbers otherwise the
// usual pseudo-random number generator.
function MDERandomPerCent(fixed) {
  this.fixed = fixed;
  this.current = 0;
  //********************
  // get repeatable random numbers
  this.random = function() {
    if (this.fixed)
      return this.nums[this.current++ % this.nums.length];
    else return Math.random() * 100.0;
  }
  //*******************
  this.reset = function(fixed) {
    if (!fixed)
      fixed = 0;
    this.current = 0;
    this.fixed = fixed;
  }
  // fixed sequence of random floats from 0-100
  this.nums =
  [ 36.265556, 18.851715, 7.3990345, 31.782913, 38.709164, 36.69688, 89.82306,
    33.963306, 48.360157, 12.216133, 85.18713, 31.195038, 24.212938, 81.62583,
    4.53164, 13.8708, 12.931156, 6.880641, 29.721706, 80.071304, 87.98213,
    57.66379, 53.464188, 64.85258, 33.507736, 87.53263, 98.40389, 50.723396,
    69.10121, 54.651283, 7.8580084, 53.125797, 59.058212, 29.690498, 24.09178,
    61.03046, 10.815275, 51.446598, 47.848106, 39.842342, 91.09411, 15.956926,
    17.054522, 32.13591, 34.74722, 92.30366, 33.573627, 95.426056, 46.163136,
    96.52513, 38.398464, 21.225452, 32.953728, 91.478836, 48.324097, 76.44714,
    4.721284, 57.694065, 12.409412, 30.220097, 3.5755575, 61.210804, 72.62313,
    94.94096, 25.377136, 88.5766, 3.2247663, 2.1845043, 50.287174, 1.5432477,
    92.07675, 63.341667, 11.884064, 95.19211, 77.90404, 56.685925, 68.848625,
    52.751396, 4.185027, 19.33552, 49.385307, 53.214657, 5.9096813, 22.160393,
    2.278161, 99.86253, 61.647606, 16.182678, 64.92522, 44.61761, 40.082787,
    61.822487, 80.12843, 45.298798, 99.36849, 36.92614, 55.87565, 19.653898,
    47.19962, 99.929375, 81.382034, 84.716446, 89.33385, 93.86219, 26.90072,
    55.669888, 49.742775, 63.007366, 11.896401, 24.67611, 96.18102, 95.15679,
    56.974667, 88.12228, 77.29585, 36.610485, 10.883331, 62.35267, 25.718433,
    11.670965, 67.57337, 29.52584, 68.64301, 21.880548, 12.239897, 60.029514,
    26.634407, 95.207825, 95.727104, 8.592671, 3.648585, 22.830147, 78.02461,
    51.12687, 78.63539, 44.78736, 74.43917, 56.783783, 99.54418, 5.073917,
    26.79525, 82.20898, 42.562622, 70.7205, 47.949837, 57.643993, 86.93876,
    23.332441, 48.580616, 10.977453, 63.742924, 34.546917, 25.898891, 55.84923,
    61.23952, 75.782845, 52.300514, 50.027214, 36.967587, 94.90292, 2.5204062,
    2.551043, 60.733437, 29.268955, 95.82522, 84.797, 74.33217, 78.46108,
    88.72304, 64.14037, 49.712772, 83.9916, 94.90849, 69.42925, 48.0983,
    71.466125, 57.225945, 76.33869, 46.79555, 64.60512, 49.957676, 80.970215,
    4.9255967, 31.804419, 62.858845, 95.683716, 16.532314, 29.28735, 16.295124,
    45.711018, 73.93521, 90.13785, 27.64228, 28.995598, 56.96834, 70.14338,
    71.91418, 78.65406, 70.1281, 57.683968, 21.530426, 90.493355, 8.140028,
    12.239599, 25.09377, 96.14293, 72.603615, 24.844206, 57.058144, 67.86322,
    97.842255, 38.948566, 32.916378, 90.276215, 50.27179, 75.180855, 13.225389,
    54.817604, 12.537062, 40.928596, 25.72109, 75.63048, 79.68318, 97.906456,
    1.592797, 51.81876, 30.683184, 48.37938, 70.1169, 19.919468, 16.67403,
    44.45604, 38.289368, 76.47456, 2.812779, 67.6492, 63.21618, 75.46925,
    6.5452757, 62.025475, 92.30386, 57.489735, 0.019699335, 13.069195, 49.8444,
    97.14293, 51.96225, 73.35356, 95.4696, 72.60651, 24.3531, 20.020437, 72.85,
    37.418533, 58.102673, 58.25147, 82.14523, 8.68944, 26.439827, 90.553474,
    32.735928, 93.23379, 85.64791, 69.995514, 9.05655, 51.300896, 21.694082,
    92.671135, 45.733185, 69.511116, 14.535332, 2.4519265, 1.4184654, 37.84334,
    23.97989, 53.361343, 93.74087, 36.21537, 60.187572, 59.022926, 64.93897,
    79.679436, 10.262895, 10.537314, 46.36846, 26.649385, 83.72045, 34.711987,
    70.355385, 29.967339, 17.981417, 92.60171, 30.91144, 33.852303, 58.900364,
    86.99658, 78.0036, 78.686134, 26.482498, 72.75908, 11.652595, 91.750175,
    32.48753, 89.23747, 17.757929, 70.63105, 44.131382, 31.820244, 9.69385,
    20.005535, 9.75709, 31.761545, 85.16655, 79.963615, 62.81949, 5.705923,
    39.484585, 61.05606, 49.5231, 61.80942, 35.73317, 85.63533, 77.51775,
    25.49802, 25.933193, 9.106332, 67.51512, 41.200645, 19.20364, 14.451528,
    58.414604, 28.033215, 52.321045, 82.21122, 94.22426, 88.03599, 48.643524,
    21.890772, 0.37084222, 1.5318096, 52.323544, 96.877914, 56.92371, 65.31709,
    39.22845, 91.5636, 51.522095, 23.197824, 61.909073, 51.366848, 31.953209,
    51.221992, 62.141674, 41.468513, 95.271614, 39.067394, 13.392716, 93.60072,
    40.04777, 89.8863, 70.74562, 22.424484, 73.39057, 12.417513, 22.809935,
    6.3175735, 59.062565, 26.877476, 1.8361866, 39.825542, 49.52506, 26.660044,
    80.38799, 29.758305, 41.720726, 16.928936, 51.257526, 5.3278923, 96.52766,
    45.951252, 80.05092, 68.61843, 43.178295, 85.457, 67.044586, 1.8218219,
    51.13021, 79.168335, 42.353966, 24.845135, 47.381298, 42.021294, 94.74877,
    44.885086, 52.07247, 86.42726, 37.727146, 54.71363, 30.76126, 94.3791,
    49.457294, 36.572754, 96.47452, 55.688198, 32.65474, 26.475393, 84.759125,
    82.92042, 49.12565, 13.885534, 36.908657, 54.39325, 80.46684, 10.226536,
    18.702293, 53.63547, 2.7680933, 21.269005, 14.89231, 14.092368, 1.6416252,
    15.710962, 59.90092, 26.5934, 55.727543, 94.78123, 99.18775, 91.37953,
    87.898705, 35.85628, 89.82987, 70.206245, 8.640802, 7.8344464, 53.420925,
    44.73924, 75.05657, 41.21629, 90.937164, 87.18669, 58.201622, 22.19994,
    31.190289, 33.916492, 55.2838, 88.797325, 23.400295, 70.43611, 2.4913013,
    32.271385, 55.191113, 82.82652, 82.71043, 34.578632, 0.89194775, 45.90522,
    25.650333, 24.910814, 17.71477, 25.646788, 28.23767, 32.903854, 63.16595,
    61.665695, 78.63714, 96.663666, 2.9212892, 23.370802, 20.539808, 49.74926,
    27.958149, 54.55992, 5.4016232, 99.62243, 63.503647, 38.909252, 48.01217,
    70.63243, 11.437541, 4.637444, 69.354515, 15.723318, 2.5553405, 89.92002,
    5.041772, 80.38764, 60.290615, 53.005486, 21.773125, 24.182188, 18.95963,
    86.8669, 76.12183, 73.78028, 17.334372, 91.96689, 74.75171, 75.99794,
    88.15369, 94.32937, 68.51214, 76.228165, 23.000294, 67.36955, 13.377571,
    85.81129, 44.634663, 99.707886, 46.93042, 92.22507, 54.110485, 8.288556,
    58.675587, 13.140577, 90.37938, 60.142483, 32.444805, 81.01418, 58.02979,
    23.404354, 89.052216, 9.643573, 51.136906, 48.480732, 22.080564, 38.571938,
    52.089752, 48.093678, 84.03856, 28.819584, 14.323484, 43.7137, 74.88837,
    94.14576, 25.231308, 96.06152, 5.2955446, 5.479807, 16.277515, 82.89659,
    60.12072, 37.498158, 74.49682, 95.55924, 39.672134, 71.406265, 67.78422,
    79.56321, 80.13348, 63.13337, 98.91243, 13.485074, 74.35457, 93.14966,
    46.663708, 62.64104, 21.736824, 16.573507, 27.371323, 34.975742, 52.026802,
    31.547081, 47.83625, 18.641579, 23.317284, 71.135895, 61.950706, 4.5190516,
    53.045547, 12.962132, 68.79762, 70.06868, 0.45318604, 74.46307, 71.561584,
    16.215252, 99.59113, 9.494495, 83.97008, 53.680252, 47.805393, 17.441643,
    37.58269, 7.262337, 47.42168, 58.894245, 74.14946, 99.52994, 50.274723,
    33.844955, 68.467186, 0.24501681, 29.690218, 99.709206, 72.06332, 25.97461,
    59.60228, 54.712116, 36.492573, 25.644934, 66.44753, 82.16396, 64.817825,
    23.814117, 12.488354, 41.226883, 5.218935, 97.47853, 34.846542, 16.122705,
    90.1739, 61.87568, 99.05559, 90.4885, 93.084946, 75.83647, 79.3604, 81.047,
    81.28399, 36.72127, 56.27473, 26.062143, 38.680584, 86.287155, 17.235147,
    35.094738, 57.838406, 53.18296, 68.85809, 30.873299, 19.056011, 22.598272,
    2.0744503, 16.749418, 73.66766, 67.456955, 27.137768, 14.259762, 29.483139,
    55.647446, 83.170166, 1.1112869, 34.654247, 22.856653, 0.057035685,
    47.399426, 93.41479, 99.57542, 93.57935, 3.5984755, 78.68476, 31.16976,
    43.685764, 40.385067, 17.888433, 29.557837, 47.631622, 52.355873, 28.92192,
    76.52008, 18.468815, 13.030881, 23.183983, 83.09362, 95.806145, 17.589642,
    26.045687, 72.34304, 45.48117, 1.6077876, 11.639935, 73.29991, 44.683777,
    16.348446, 59.54288, 57.809036, 98.18703, 20.038408, 68.23101, 65.193275,
    60.18984, 22.80209, 81.20797, 43.352467, 66.01515, 83.7643, 3.2835484,
    52.271248, 97.11268, 41.904007, 8.312869, 35.84886, 39.4456, 31.76502,
    22.280062, 37.654694, 14.592057, 62.429935, 25.32134, 13.142073, 0.1294195,
    84.398285, 95.848724, 65.07574, 14.932227, 21.72035, 94.10121, 1.7994285,
    35.093666, 96.1051, 79.8144, 65.24356, 11.754858, 82.3412, 16.35217, 42.3,
    69.51981, 71.654945, 51.859444, 69.030624, 25.485659, 32.354034, 5.186015,
    56.142414, 35.053974, 82.10312, 93.84582, 41.00756, 77.80948, 59.02795,
    88.15528, 87.64535, 93.08067, 40.699978, 14.96147, 56.83772, 19.146109,
    75.680695, 6.0685277, 0.61359406, 80.3352, 37.668507, 3.0752957, 47.507935,
    48.21102, 88.479385, 36.77013, 28.781378, 48.055767, 82.065186, 70.57084,
    85.7401, 42.19342, 71.12895, 5.784088, 62.364716, 93.07161, 44.250275,
    54.804276, 58.094143, 68.293884, 96.10559, 9.087581, 44.968475, 62.10609,
    34.04197, 62.09839, 48.3737, 19.512642, 3.5823405, 42.40019, 9.865337,
    97.28324, 49.144535, 11.834168, 53.849922, 66.01912, 38.5302, 12.209135,
    33.54984, 95.39827, 22.79616, 73.78669, 37.47777, 28.255154, 3.8264513,
    70.14557, 32.21743, 10.05168, 95.77759, 72.153534, 80.39812, 18.359262,
    14.805412, 33.53851, 10.13707, 97.663574, 8.561409, 5.9840083, 4.9827156,
    41.1333, 59.236042, 27.274805, 85.058685, 75.62566, 65.74114, 73.47133,
    87.84579, 69.23384, 16.779024, 40.626335, 96.848595, 5.7576475, 31.913233,
    96.460464, 95.33925, 15.361542, 87.87564, 87.13171, 96.2557, 13.928193,
    46.060707, 1.7482698, 71.98914, 39.46245, 14.1602335, 46.195675, 75.729385,
    93.045525, 26.727343, 7.7908096, 4.764533, 15.500223, 54.90055, 97.48827,
    10.1077795, 90.69149, 87.413475, 9.456289, 95.55761, 37.45395, 20.721363,
    12.324071, 33.431072, 14.607596, 3.2046795, 42.843132, 85.33219, 35.644318,
    15.631569, 82.58396, 87.60254, 4.0392876, 82.31862, 12.725491, 79.73848,
    60.77274, 54.725204, 91.43609, 96.8527, 80.701675, 80.73489, 8.053088,
    84.19302, 75.31121, 13.870573, 74.88167, 31.029236, 10.688764, 8.703566,
    78.15937, 15.179682, 6.7450166, 52.512444, 45.94433, 96.64514, 2.9709697,
    86.89364, 0.07891655, 89.67096, 61.804207, 64.625916, 6.842375, 43.28727,
    5.287266, 11.44976, 55.86216, 16.028547, 98.65484, 21.394045, 7.61708,
    37.3184, 46.35877, 96.05226, 99.15082, 77.27823, 96.494286, 15.058035,
    72.11543, 64.42676, 27.889866, 63.55989, 88.5667, 54.80599, 65.477516,
    68.03942, 62.956314, 15.791923, 58.004658, 67.1159, 42.61073, 47.814167,
    42.413998, 54.24544, 20.641846, 11.034173, 95.958916, 84.07739, 59.45096,
    30.891197, 79.64882, 79.81116, 66.70736, 48.120945, 83.24208, 59.69662,
    26.25386, 95.6711, 13.068247, 88.33003, 96.296, 72.796646, 40.63471,
    88.354065, 26.394474, 14.164686, 63.269077, 71.14305, 24.722797, 30.131895,
    59.75948, 15.848893, 51.93436, 39.23676, 61.491226, 91.33394, 83.07977,
    94.159355, 5.9750137, 96.91013, 76.842316, 50.476734, 96.21457, 25.701672,
    9.858221, 81.022484, 38.57108, 85.54808, 28.05475, 86.97887, 44.970947,
    1.9936204, 39.84322, 12.261211, 53.15857, 7.3747873, 64.31908, 28.583849,
    14.570308, 22.71452, 6.504667, 63.86115, 37.39761, 1.6929448, 27.434475,
    42.028748, 38.75081, 57.14304, 91.532104, 25.482607, 14.171582, 21.72404,
    33.673904, 77.069046, 24.010695, 22.123224, 31.232029, 16.540045, 34.79127,
    13.3867445, 93.08369, 56.133194, 65.579834, 0.25707483, 34.321434, 55.4921,
    15.325022, 77.135704, 86.50119, 87.59048 ]
}
///////////////////////////////////////////////////////////////////////////////
// EOF edimpro1.js
